/*

Mures
Copyright (C) 2001 Adam D'Angelo

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact information:

Adam D'Angelo
dangelo@ntplx.net
P.O. Box 1155
Redding, CT 06875-1155
USA

*/

#include <stdlib.h>
#include "SDL.h"

#include "protocol.h"
#include "sim.h"
#include "input.h"
#include "keyboard.h"
#include "output.h"
#include "lib.h"
#include "main.h"

typedef enum _input_action_type {
  POINTER_ACCEL,
  POINTER_DECEL,
  POINTER_MOVE,
  PLACE_ARROW
} input_action_type;

typedef struct _input_action {
  int exists;
  int player;
  input_action_type type;
  direction dir;
} input_action;

input_action key_up_action[SDLK_EURO+1];
input_action key_down_action[SDLK_EURO+1];

void keyboard_data_clear(keyboard_data *data)
{
  int i;
  
  for(i=0; i<MAX_PLAYER; i++) {
    data->player[i].exists = 0;

    data->player[i].pointer.x = 0;
    data->player[i].pointer.y = 0;
    data->player[i].pointer_going.x = 0;
    data->player[i].pointer_going.y = 0;
    data->player[i].on_grid = 0;
  }
}

void handle_action(sim *s, keyboard_data *data, input_action action, input_queue *q)
{
  if(action.player == -1)
    return;

  if(!data->player[action.player].exists)
    return;

  switch(action.type) {
  case PLACE_ARROW:
    if(data->player[action.player].on_grid)
      if(free_for_arrow(s, data->player[action.player].grid_pos))
	input_enqueue(q, action.player, action.dir, data->player[action.player].grid_pos);
    break;
  case POINTER_DECEL:
    switch(action.dir) {
    case up: data->player[action.player].pointer_going.y++; break;
    case down: data->player[action.player].pointer_going.y--; break;
    case left: data->player[action.player].pointer_going.x++; break;
    case right: data->player[action.player].pointer_going.x--; break;
    default: break;
    }
    break;
  case POINTER_ACCEL:
    switch(action.dir) {
    case up: data->player[action.player].pointer_going.y--; break;
    case down: data->player[action.player].pointer_going.y++; break;
    case left: data->player[action.player].pointer_going.x--; break;
    case right: data->player[action.player].pointer_going.x++; break;
    default: break;
    }
    break;
  default:
    fprintf(stderr, "Unknown action type: %d\n", action.type);
    break;
  }

  return;
}

void set_action(SDLKey key, int player, input_action_type type, direction dir)
{
  key_down_action[key].exists = 1;
  key_down_action[key].player = player;
  key_down_action[key].type = type;
  key_down_action[key].dir = dir;

  if(type==POINTER_ACCEL) {
    key_up_action[key] = key_down_action[key];
    key_up_action[key].type = POINTER_DECEL;
  }
}

void keyboard_new_map(sim *s, keyboard_data *data, int use_mouse)
{
  int i;
  screen_position p;

  keyboard_init(use_mouse);

  for(i=0; i<MAX_PLAYER; i++) {
    if(have_local_player[i]) {
      if(local_player_type[i] == KEYBOARD_PLAYER)
	data->player[i].exists = 1;
    }
    else
      data->player[i].exists = 0;
  }

  for(i=0; i<MAX_PLAYER; i++)
    if(data->player[i].exists) {
      p = grid2screen(s->player[i].rocket);
      
      p.x += BLOCK_SIZE/2;
      p.y += BLOCK_SIZE/2;
      data->player[i].pointer = p;

      printf("p is %d,%d\n", p.x, p.y);
      
      if(i==local_keyboard_player(0))
	if(use_mouse)
	  SDL_WarpMouse(p.x/MOUSE_INPUT_SPEED, p.y/MOUSE_INPUT_SPEED);
    }
}


void keyboard_init(int use_mouse)
{
  /* player 1 */

  set_action(SDLK_KP8, local_keyboard_player(1), PLACE_ARROW, up);
  set_action(SDLK_KP2, local_keyboard_player(1), PLACE_ARROW, down);
  set_action(SDLK_KP5, local_keyboard_player(1), PLACE_ARROW, down);
  set_action(SDLK_KP6, local_keyboard_player(1), PLACE_ARROW, right);
  set_action(SDLK_KP4, local_keyboard_player(1), PLACE_ARROW, left);

  set_action(SDLK_UP, local_keyboard_player(1), POINTER_ACCEL, up);
  set_action(SDLK_DOWN, local_keyboard_player(1), POINTER_ACCEL, down);
  set_action(SDLK_LEFT, local_keyboard_player(1), POINTER_ACCEL, left);
  set_action(SDLK_RIGHT, local_keyboard_player(1), POINTER_ACCEL, right);

  /* player 2 */

  set_action(SDLK_j, local_keyboard_player(2), PLACE_ARROW, left);
  set_action(SDLK_k, local_keyboard_player(2), PLACE_ARROW, down);
  set_action(SDLK_l, local_keyboard_player(2), PLACE_ARROW, right);
  set_action(SDLK_i, local_keyboard_player(2), PLACE_ARROW, up);

  set_action(SDLK_a, local_keyboard_player(2), POINTER_ACCEL, left);
  set_action(SDLK_s, local_keyboard_player(2), POINTER_ACCEL, down);
  set_action(SDLK_d, local_keyboard_player(2), POINTER_ACCEL, right);
  set_action(SDLK_w, local_keyboard_player(2), POINTER_ACCEL, up);

}

int keyboard_step(sim *s, input_queue *q, keyboard_data *data, int use_output, int ms)
{
  static screen_position from;

  SDL_Event event;
  int x,y,i,found;
  direction dir;
  int p;

  if(use_output)
    for(i=0; i<MAX_PLAYER; i++) {

      data->player[i].pointer.x += data->player[i].pointer_going.x*ms*POINTER_SPEED;
      data->player[i].pointer.y += data->player[i].pointer_going.y*ms*POINTER_SPEED;
      if(data->player[i].pointer.x > screen->w)
	data->player[i].pointer.x = screen->w;
      if(data->player[i].pointer.y > screen->h)
	data->player[i].pointer.y = screen->h;
      if(data->player[i].pointer.x < 0)
	data->player[i].pointer.x = 0;
      if(data->player[i].pointer.y < 0)
	data->player[i].pointer.y = 0;

      data->player[i].on_grid = on_grid(data->player[i].pointer);
      if(data->player[i].on_grid)
	data->player[i].grid_pos = screen2grid(data->player[i].pointer);
    }

  while(SDL_PollEvent(&event)) {
    if(event.type == SDL_QUIT)
      return 0;
    
    switch(event.type) {
    case SDL_VIDEORESIZE:
      change_screen(event.resize.w, event.resize.h);
      break;
    case SDL_MOUSEBUTTONDOWN:
      if(local_keyboard_player(0) >= 0)
	from = data->player[local_keyboard_player(0)].pointer;
      break;
    case SDL_MOUSEMOTION:
      if(local_keyboard_player(0) >= 0) {
	data->player[local_keyboard_player(0)].pointer.x = event.motion.x*MOUSE_INPUT_SPEED;
	data->player[local_keyboard_player(0)].pointer.y = event.motion.y*MOUSE_INPUT_SPEED;
	data->player[local_keyboard_player(0)].on_grid = on_grid(data->player[local_keyboard_player(0)].pointer);
	if(data->player[local_keyboard_player(0)].on_grid)
	  data->player[local_keyboard_player(0)].grid_pos = screen2grid(data->player[local_keyboard_player(0)].pointer);
      }
      break;
    case SDL_MOUSEBUTTONUP:
      if(local_keyboard_player(0) >= 0) {
	switch(event.button.button) {
	case SDL_BUTTON_LEFT: p = local_keyboard_player(0); break;
	case SDL_BUTTON_RIGHT:
	  if(on_grid(data->player[local_keyboard_player(0)].pointer))
	    if(free_for_arrow(s, screen2grid(data->player[local_keyboard_player(0)].pointer)))
	      input_enqueue(q, local_keyboard_player(0), right, screen2grid(data->player[local_keyboard_player(0)].pointer));
	  p = -1;
	  break;
	default: p = -1; break;
	}
	if(p>=0) {
	  x = data->player[local_keyboard_player(0)].pointer.x;
	  y = data->player[local_keyboard_player(0)].pointer.y;
	  
	  if(abs(from.x-x) > abs(from.y-y)) {
	    if(x>from.x)
	      dir = right;
	    else
	      dir = left;
	  }
	  else {
	    if(y>from.y)
	      dir = down;
	    else
	      dir = up;
	  }
	  if(on_grid(from))
	    if(free_for_arrow(s, screen2grid(from)))
	      input_enqueue(q, p, dir, screen2grid(from));
	}
      }
      break;
    case SDL_KEYUP:
      if(key_up_action[event.key.keysym.sym].exists)
	handle_action(s, data, key_up_action[event.key.keysym.sym], q);
      break;
    case SDL_KEYDOWN:
      if(key_down_action[event.key.keysym.sym].exists)
	handle_action(s, data, key_down_action[event.key.keysym.sym], q);
      else
	switch(event.key.keysym.sym) {
	case SDLK_h:
	  if(data->player[0].on_grid)
	    s->map.hwall[data->player[0].grid_pos.x][data->player[0].grid_pos.y] = s->map.hwall[data->player[0].grid_pos.x][data->player[0].grid_pos.y]?0:1;
	  break;
	case SDLK_v:
	  if(data->player[0].on_grid)
	    s->map.vwall[data->player[0].grid_pos.x][data->player[0].grid_pos.y] = s->map.vwall[data->player[0].grid_pos.x][data->player[0].grid_pos.y]?0:1;
	  break;
	case SDLK_o:
	  if(data->player[0].on_grid)
	    s->map.hole[data->player[0].grid_pos.x][data->player[0].grid_pos.y] = s->map.hole[data->player[0].grid_pos.x][data->player[0].grid_pos.y]?0:1;
	  break;
	case SDLK_r:
	  if(data->player[0].on_grid)
	    s->player[0].rocket = data->player[0].grid_pos;
	  break;
	case SDLK_g:
	  found = 0;
	  for(i=0; i<MAX_GENERATOR; i++)
	    if(s->generator[i].exists)
	      if(s->generator[i].pos.x == data->player[0].grid_pos.x && s->generator[i].pos.y == data->player[0].grid_pos.y) {
		s->generator[i].exists = 0;
		found = 1;
	      }
	  if(!found)
	    add_generator(s, data->player[0].grid_pos, rand()%4);
	  break;
	case SDLK_PRINT:
	  output_screenshot();
	  break;
	case SDLK_p:
	case SDLK_BREAK:
	  if(s->authority)
	    sim_toggle_pause(s);
	  break;
	case SDLK_RETURN:
	  if(s->authority)
	    sim_start(s);
	  break;
/*  	case SDLK_s: */
/*  	  save_map("saved.chu"); */
/*  	  break; */
	case SDLK_ESCAPE:
	case SDLK_q:
	  return 0;
	  break;
	default:
	  break;
	}
      break;
    }
  }

  return 1;
}  


