/*

Mures
Copyright (C) 2001 Adam D'Angelo

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact information:

Adam D'Angelo
dangelo@ntplx.net
P.O. Box 1155
Redding, CT 06875-1155
USA

*/

#include <stdio.h>
#include <string.h>
#include "protocol.h"
#include "sim.h"
#include "lib.h"
#include "map.h"
#include "main.h"

#define MAX_VERSION 1.0

char *default_file = "default.chu";

void clear_map(map *m)
{
  int i, j;

  m->type = -1;
  m->max_player = 0;
  m->max_generator = 0;
  m->max_creature = 0;

  for(i=0; i<NUM_BLOCKS_X+1; i++)
    for(j=0; j<NUM_BLOCKS_Y+1; j++)
      m->hwall[i][j] = m->vwall[i][j] = 0;
  
  for(i=0; i<NUM_BLOCKS_X; i++)
    m->hwall[i][0] = m->hwall[i][NUM_BLOCKS_Y] = 1;
  for(i=0; i<NUM_BLOCKS_Y; i++)
    m->vwall[0][i] = m->vwall[NUM_BLOCKS_X][i] = 1;
  
  for(i=0; i<NUM_BLOCKS_X; i++)
    for(j=0; j<NUM_BLOCKS_Y; j++)
      m->hole[i][j] = 0;
}

int load_map(map *m, char *filename)
{
  char buff[MAX_STRING], temp;
  char buff2[MAX_STRING], buff3[MAX_STRING];
  char *line;
  int p,i;
  int got_version = 0, got_type = 0;
  int data_mode = 0;
  int data_line = 0;
  float version;
  FILE *in = fopen(filename, "r");

  printf("Loading map %s...\n", filename);

  if(in==NULL) {
    fprintf(stderr, "Error reading file \"%s\".\n", filename);
    return 0;
  }

  clear_map(m);
  
  p=0;

  while(!feof(in)) { /* for each line */

    /* put line into buff */
    i=0;
    while(1) {
      if(feof(in) || i==MAX_STRING)
	break;

      fscanf(in, "%c", &temp);

      if(temp == '\n')
	break;

      buff[i++] = temp;
    }

    /* trim trailing whitespace */
    while(i>0 && buff[i-1]==' ')
      i--;
    buff[i]=0;
    
    line = &buff[0];
    
    while(*line==' ')
      line++;
    
    if(*line==0 || *line=='#')
      continue;
    
    strcpy(&buff2[0], line);

    if(!data_mode) {
      i=0;
      while(buff2[i]!=' ' && buff2[i]!=0)
	i++;

      if(buff2[i]==0) /* no space in line */
	buff3[0] = 0;
      else {
	buff2[i] = 0;
	strcpy(&buff3[0], &buff2[i+1]);
      }

      i=0;
      while(buff2[i]) {
	if(buff2[i]>='A' && buff2[i]<='Z')
	  buff2[i] = buff2[i] - 'A' + 'a';
	i++;
      }
      
      if(!strcmp(buff2, "version")) {
	got_version = 1;
	sscanf(buff3, "%f", &version);
	if(version > MAX_VERSION) {
	  fprintf(stderr, "Map version %.2f not supported.\n", version);
	  return 0;
	}
      }
      
      if(!strcmp(buff2, "type")) {
	got_type = 1;
	i=0;
	while(buff3[i]) {
	  if(buff3[i]>='A' && buff3[i]<='Z')
	    buff3[i] = buff3[i] - 'A' + 'a';
	  i++;
	}
	m->type = 200;
	if(!strcmp(buff3, "battle"))
	  m->type = BATTLE;
	if(m->type == 200) {
	  fprintf(stderr, "Unsupported map type: %s\n", buff3);
	  return 0;
	}
      }

      if(!strcmp(buff2, "data")) {
	if(got_version)
	  data_mode = 1;
	else {
	  fprintf(stderr, "Data started without valid version.\n");
	  return 0;
	}
      }
    }
    else {
      if(data_line%2==0) { /* lines of horizontal walls */
	if(data_line/2 > NUM_BLOCKS_Y+1)
	  fprintf(stderr, "Too many data lines in file.\n");
	else
	  for(i=0; i<NUM_BLOCKS_X; i++)
	    if(buff2[i*2+1]!=' ')
	      m->hwall[i][data_line/2] = 1;
      }
      else { /* lines of vertical walls, holes, more */
	if((data_line-1)/2 > NUM_BLOCKS_Y+1)
	  fprintf(stderr, "Too many data lines in file.\n");
	else {
	  /* check vwalls */
	  for(i=0; i<NUM_BLOCKS_X+1; i++) {
	    if(buff2[i*2]!=' ')
	      m->vwall[i][(data_line-1)/2] = 1;
	  }

	  for(i=0; i<NUM_BLOCKS_X; i++) {
	    if(buff2[i*2+1] == 'P') {
	      if(m->max_player < MAX_PLAYER) {
		m->player_start[m->max_player].x = i;
		m->player_start[m->max_player].y = (data_line-1)/2;
		m->max_player++;
	      }
	    }
	    else if(buff2[i*2+1]=='O')
	      m->hole[i][(data_line-1)/2] = 1;
	    else if(char_to_dir(buff2[i*2+1]) < MAX_DIR) {
	      if(m->max_generator < MAX_GENERATOR) {
		m->generator[m->max_generator].pos = grid_int_pos(i, (data_line-1)/2);
		m->generator[m->max_generator].dir = char_to_dir(buff2[i*2+1]);
		m->max_generator++;
	      }
	    }
	  }
	}
      }
	

      data_line++;
    }
  }
  
  fclose(in);
  
  if(!data_mode) {
    fprintf(stderr, "Map contains no data.\n");
    return 0;
  }

  /* for safety */

  return 1;
}

int save_map(sim s, char *location)
{
  int x,y,i,done;
  FILE *out = fopen(location, "w");
  if(out==NULL) {
    fprintf(stderr, "Error writing file \"%s\".\n", location);
    return 0;
  }
  fprintf(out, "#Mures map\n");
  fprintf(out, "Version %.2f\n", MAX_VERSION);
  fprintf(out, "Type %s\n", map_type_to_s(s.map.type));
  fprintf(out, "Data\n");

  for(y=0; y<NUM_BLOCKS_Y+1; y++) {
    for(x=0; x<NUM_BLOCKS_X+1; x++) {
      fprintf(out, "+");
      if(x<NUM_BLOCKS_X)
	fprintf(out, s.map.hwall[x][y]?"-":" ");
    }
    fprintf(out, "\n");

    if(y<NUM_BLOCKS_Y) {
      for(x=0; x<NUM_BLOCKS_X+1; x++) {
	fprintf(out, s.map.vwall[x][y]?"|":" ");
	if(x<NUM_BLOCKS_X) {
	  done = 0;
	  for(i=0; i<MAX_PLAYER; i++)
	    if(s.player[i].exists)
	      if(s.player[i].rocket.x == x && s.player[i].rocket.y == y) {
		fprintf(out, "%d", i+1);
		done = 1;
		break;
	      }
	  for(i=0; i<MAX_GENERATOR; i++)
	    if(s.generator[i].exists)
	      if(s.generator[i].pos.x == x && s.generator[i].pos.y == y) {
		fprintf(out, "%c", dir_to_char(s.generator[i].dir));
		done = 1;
		break;
	      }
	  if(!done)
	    fprintf(out, s.map.hole[x][y]?"O":" ");
	}
      }
      fprintf(out, "\n");
    }
  }
  return 1;
}

