/*

Mures
Copyright (C) 2001 Adam D'Angelo

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact information:

Adam D'Angelo
dangelo@ntplx.net
P.O. Box 1155
Redding, CT 06875-1155
USA

*/

#include <stdlib.h>
#include <math.h>
#include "SDL.h"
#include "SDL_image.h"
#include "SFont.h"
#include "protocol.h"
#include "sim.h"
#include "output.h"
#include "input.h"
#include "keyboard.h"
#include "audio.h"

SDL_Color grid_box_color = {0xFF, 0xFF, 0xFF, 0};
SDL_Color grid_box_grey_color = {0x88, 0x88, 0x88, 0};
SDL_Color fence_color = { 0x00, 0x00, 0xAA, 0 };
SDL_Color fence_post_color = { 0x00, 0x00, 0xAA, 0};
SDL_Color white = { 0xFF, 0xFF, 0xFF, 0 };
SDL_Color red = {0xFF, 0, 0, 0};
SDL_Color black = { 0x00, 0x00, 0x00, 0 };
SDL_Color yellow = { 0xFF, 0xDD, 0x88, 0 };
SDL_Color blue = { 0x88, 0x88, 0xCC, 0 };
SDL_Color blue2 = { 0x66, 0x66, 0x99, 0 };
SDL_Color arrow_color = {0xFF, 0xFF, 0xFF};
SDL_Color box_border_color = {0xAA, 0xAA, 0xAA};
SDL_Color box_color = {0x33, 0x33, 0x33};

#define DARK 0x33
#define LIGHT 0xFF

SDL_Color player_color[MAX_PLAYER] = {
  {DARK, DARK, LIGHT}, /* blue   */
  {LIGHT*9/10, LIGHT*8/10, DARK},  /* yellow */
  {LIGHT, DARK, DARK}, /* red    */
  {DARK, LIGHT*9/10, DARK} /* green  */
};

#define MAX_IMAGE 24

SDL_Surface *background = NULL;
SDL_Surface *image[MAX_IMAGE];
SDL_Surface *font = NULL;

int have_bg;

typedef enum _image_id
{
  MOUSE,
  MOUSE_Q,
  MOUSE_50,
  ROCKET,
  ROCKET_1,
  ROCKET_2,
  ROCKET_3,
  POINTER,
  POINTER_1,
  POINTER_2,
  POINTER_3,
  WALLPOST,
  CAT,
  GRID_BOX,
  GRID_BOX_GREY,
  GENERATOR,
  PLACE_1,
  PLACE_2,
  PLACE_3,
  PLACE_4
} image_id;

#define NUM_DIR 4

SDL_Surface *arrow_image[MAX_PLAYER][NUM_DIR][ARROW_TOUGHNESS+1];
SDL_Surface *digit_image[12];

#define PLUS 10
#define MINUS 11

typedef struct _image_info_type
{
  image_id id;
  char *file;
  int trans;
  int alpha;
} image_info_type;

image_info_type image_info[MAX_IMAGE] = {
  { MOUSE, "mouse.gif", 1, 0 },
  { MOUSE_Q, "mouseq.gif", 1, 0 },
  { MOUSE_50, "mouse50.gif", 1, 0 },
  { ROCKET, "rocket0.gif", 1, 0 },
  { ROCKET_1, "rocket1.gif", 1, 0 },
  { ROCKET_2, "rocket2.gif", 1, 0 },
  { ROCKET_3, "rocket3.gif", 1, 0 },
  { POINTER, "pointer0.gif", 1, 0 },
  { POINTER_1, "pointer1.gif", 1, 0},
  { POINTER_2, "pointer2.gif", 1, 0},
  { POINTER_3, "pointer3.gif", 1, 0},
  { CAT, "cat.gif", 1, 0 },
  { GENERATOR, "generator.gif", 1, 0 },
  { PLACE_1, "place1.gif", 1, 0},
  { PLACE_2, "place2.gif", 1, 0},
  { PLACE_3, "place3.gif", 1, 0},
  { PLACE_4, "place4.gif", 1, 0}
};

int player_rank[MAX_PLAYER];

void output_handle_event(int event)
{
  switch(event) {
  case GET_MOUSE:
    play_sound("sounds/get_mouse.wav");
    break;
  case GET_MOUSE_50:
    play_sound("sounds/get_mouse_50.wav");
    break;
  case GET_MOUSE_Q:
    play_sound("sounds/get_mouse_q.wav");
    break;
  case GET_CAT:
    play_sound("sounds/get_cat.wav");
    break;
  case CLOCK_TICK:
    play_sound("sounds/clock_tick.wav");
    break;
  case START_GAME:
    play_sound("sounds/start_game.wav");
    break;
  case PAUSE_GAME:
    play_sound("sounds/pause_game.wav");
    break;
  default:
    fprintf(stderr, "Couldn't handle output event of type %d\n", event);
    break;
  }
}

char *mode_string(qmouse_type mode)
{
  switch(mode) {
  case Q_MOUSE_MANIA: return "Mouse Mania!"; break;
  case Q_CAT_MANIA: return "Cat Mania!"; break;
  case Q_SPEED_UP: return "Speed Up!"; break;
  case Q_SLOW_DOWN: return "Slow Down!"; break;
  case Q_MOUSE_MONOPOLY: return "Mouse Monopoly!"; break;
  case Q_CAT_ATTACK: return "Cat Attack!"; break;
  case Q_EVERYBODY_MOVE: return "Everybody Move!"; break;
  case Q_PLACE_ARROWS_AGAIN: return "Place Arrows Again!"; break;
  default:
    fprintf(stderr, "didn't know what to say for mode %d\n", mode);
    return "error";
    break;
  }
}


SDL_Surface* creature_image(int type)
{
  switch(type) {
  case cat:
    return image[CAT];
    break;
  case mouse:
    return image[MOUSE];
    break;
  case mouse50:
    return image[MOUSE_50];
    break;
  case mouseq:
    return image[MOUSE_Q];
    break;
  default:
    fprintf(stderr, "Don't know how to render creature of type %d\n", type);
    return image[MOUSE];
    break;
  }
}
    

SDL_Surface* player_rocket_image(int player)
{
  int num;
  switch(player) {
  case 0: num = ROCKET; break;
  case 1: num = ROCKET_1; break;
  case 2: num = ROCKET_2; break;
  case 3: num = ROCKET_3; break;
  default:
    fprintf(stderr, "No rocket image for player %d.\n", player);
    exit(1);
    break;
  }
  return image[num];
}

SDL_Surface *player_pointer_image(int player)
{
  int num;
  switch(player) {
  case 0: num = POINTER; break;
  case 1: num = POINTER_1; break;
  case 2: num = POINTER_2; break;
  case 3: num = POINTER_3; break;
  default:
    fprintf(stderr, "No pointer image for player %d.\n", player);
    exit(1);
    break;
  }
  return image[num];
}

grid_int_position screen2grid(screen_position onscreen)
{
  grid_int_position temp;
  if(onscreen.x<LEFT_BORDER || onscreen.y<TOP_BORDER)
    return grid_int_pos(-1, -1);

  temp.x = (onscreen.x-LEFT_BORDER)/BLOCK_SIZE;
  temp.y = (onscreen.y-TOP_BORDER)/BLOCK_SIZE;
  if(valid_grid_int(temp))
    return temp;
  else
    return grid_int_pos(-1, -1);
}

int on_grid(screen_position onscreen)
{
  grid_int_position temp = screen2grid(onscreen);
  if(temp.x==((grid_int)-1) && temp.y==((grid_int)-1))
    return 0;
  else
    return 1;
}

screen_position grid2screen(grid_int_position ongrid)
{
  screen_position temp;

  if(screen==NULL) {
    fprintf(stderr, "Error: tried to get screen coordinates when screen wasn't initialized.\n");
    exit(1);
  }

  temp.x = ongrid.x*BLOCK_SIZE+LEFT_BORDER;
  temp.y = ongrid.y*BLOCK_SIZE+TOP_BORDER;
  return temp;
}

screen_position gridf2screen(grid_float_position ongrid)
{
  screen_position temp;
  temp.x = ongrid.x*BLOCK_SIZE+LEFT_BORDER;
  temp.y = ongrid.y*BLOCK_SIZE+TOP_BORDER;
  return temp;
}

screen_position screen_pos(Uint32 x, Uint32 y)
{
  screen_position temp;
  temp.x = x;
  temp.y = y;
  return temp;
}

void unload_images();
int load_images();

void set_mode(int width, int height, int fullscreen)
{
  int options = SDL_DOUBLEBUF | SDL_HWSURFACE;

  if(fullscreen)
    options |= SDL_FULLSCREEN;
  else
    options |= SDL_RESIZABLE;

  screen = SDL_SetVideoMode(width, height, 0, options);

  if ( screen == NULL ) {
    fprintf(stderr, "Couldn't set %dx%d video mode: %s\n", SCREEN_WIDTH, SCREEN_HEIGHT, SDL_GetError());
    exit(2);
  }
  else
    printf("Screen allocated.\n");

  unload_images();
  if(!load_images()) {
    fprintf(stderr, "Couldn't load images, quitting.\n");
    exit(1);
  }
}

void change_screen(int new_width, int new_height)
{
  set_mode(new_width, new_height, 0);
  
  printf("Changing to %dx%d\n", new_width, new_height);
  have_bg = 0;
}

SDL_Surface *LoadImage(char *datafile, int transparent)
{
	SDL_Surface *image, *surface;

	image = IMG_Load(datafile);
	if ( image == NULL ) {
		fprintf(stderr, "Couldn't load image %s: %s\n",
					datafile, IMG_GetError());
		return(NULL);
	}
	if ( transparent ) {
		/* Assuming 8-bit BMP image */
		SDL_SetColorKey(image, (SDL_SRCCOLORKEY|SDL_RLEACCEL),
						*(Uint8 *)image->pixels);
	}
	surface = SDL_DisplayFormat(image);
	SDL_FreeSurface(image);
	return(surface);
}

SDL_Surface *load_image(char *datafile, int transparent)
{
  SDL_Surface *image, *surface;

  image = IMG_Load(datafile);
  if (image == NULL) {
    fprintf(stderr, "Couldn't load image %s: %s\n",
	    datafile, IMG_GetError());
    return(NULL);
  }
  if (transparent) {
    /* Assuming 8-bit BMP image */
    SDL_SetColorKey(image, (SDL_SRCCOLORKEY|SDL_RLEACCEL), 
		    *(Uint8 *)image->pixels);
  }
  surface = SDL_DisplayFormat(image);
  SDL_FreeSurface(image);
  return(surface);
}

void ShowImage(SDL_Surface *image, int x, int y)
{
    SDL_Rect dest;

    /* Blit onto the screen surface.
       The surfaces should not be locked at this point.
     */
    dest.x = x;
    dest.y = y;
    dest.w = image->w;
    dest.h = image->h;

    if(SDL_BlitSurface(image, NULL, screen, &dest) == -2) {
      /* we have it back */
      if(SDL_LockSurface(image) > 0) {
	SDL_UnlockSurface(image);
	
	unload_images();
	load_images();
      }
    }
}

void show_image(SDL_Surface *image, screen_position p)
{
  ShowImage(image, p.x, p.y);
}

void center_image(SDL_Surface *image, screen_position top_left)
{
  screen_position pos = top_left;

  pos.x += BLOCK_SIZE/2;
  pos.y += BLOCK_SIZE/2;
  pos.x -= image->w/2;
  pos.y -= image->h/2;
  
  ShowImage(image, pos.x, pos.y);
}

void center_rocket(sim *s, int player, screen_position top_left)
{
  int i, count=0, place, right, left;
  SDL_Rect src, dest;

  for(i=0; i<MAX_PLAYER; i++)
    if(s->player[i].exists)
      if(s->player[i].team_leader == s->player[player].team_leader)
	count++;

  dest.x = top_left.x;
  dest.y = top_left.y;

  /* center */
  dest.x += BLOCK_SIZE/2;
  dest.y += BLOCK_SIZE/2;
  dest.x -= player_rocket_image(0)->w/2;
  dest.y -= player_rocket_image(0)->h/2;

  left = dest.x;

  src.y=0;

  src.h = player_rocket_image(0)->h;
  dest.h = src.h;

  for(i=0, place=0; i<MAX_PLAYER; i++)
    if(s->player[i].exists)
      if(s->player[i].team_leader == s->player[player].team_leader) {
	src.x = player_rocket_image(0)->w*place/count;
	dest.x = src.x + left;
	right = player_rocket_image(0)->w*(place+1)/count;
	dest.w = right-left;
	src.w = dest.w;
	
	SDL_BlitSurface(player_rocket_image(i), &src, screen, &dest);
	place++;
      }
}

void makebg(sim *s)
{
  SDL_Rect rect;
  int x, y;
  screen_position p;
  int odd;
  SDL_Color color;
  SDL_Color checker_one = blue2; /* yellow */
  SDL_Color checker_two = blue;

  background = SDL_CreateRGBSurface(SDL_HWSURFACE, screen->w, screen->h, 16, 0, 0, 0, 0);

  /* blackness */
  rect.x = 0;
  rect.y = 0;
  rect.w = screen->w;
  rect.h = screen->h;
  SDL_FillRect(background, &rect, SDL_MapRGB(screen->format, 0, 0, 0));

  /* checkerboard */
  odd=0;
  for(x=0; x<NUM_BLOCKS_X; x++)
    for(y=0; y<NUM_BLOCKS_Y; y++) {
      p = grid2screen(grid_int_pos(x,y));
      rect.x = p.x;
      rect.y = p.y;

      /* this needs to be set every time because SDL_FillRect will change it if it goes offscreen */
      rect.w = rect.h = BLOCK_SIZE;

      odd = (x%2+y%2)%2;
      if(odd)
	color = checker_one;
      else
	color = checker_two;
      if(s->map.hole[x][y])
	color = black;

      SDL_FillRect(background, &rect, SDL_MapRGB(screen->format, color.r, color.g, color.b));
    }

  have_bg = 1;
}

void clearbg()
{
  SDL_FreeSurface(background);
  have_bg = 0;
}

void output_new_map(sim *s)
{

  fprintf(stderr, "in output_new_map...\n");
  if(background != NULL)
    clearbg();
  makebg(s);
}

void center_text(int x, int y, char *text)
{
  PutString(screen, x-TextWidth(text)/2, y-font->h/2, text);
}  

int get_winner(sim *s)
{
  int i,max,winner;

  winner=-1;
  max=-1;

  for(i=0; i<MAX_PLAYER; i++) {
    if(s->player[i].score == max) /* tie */
      winner = -1;
    else if(s->player[i].score > max) {
      max = s->player[i].score;
      winner = i;
    }
  }

  return winner;
}


void output_refresh(sim *s, keyboard_data *data)
{
  int i,j,k,x,y=0,max_score,max_player;
  
  SDL_Rect rect;
  screen_position p;
  float per;
  int left;
  SDL_Surface *temp_image;

  if(!have_bg)
    makebg(s);

  show_image(background, screen_pos(0, 0));

  /* arrows */
  for(i=0; i<MAX_PLAYER; i++)
    if(s->player[i].exists)
      for(j=0; j<MAX_ARROW; j++)
	if(s->player[i].arrow[j].exists)
	  if(s->player[i].arrow[j].time_left > FLICKER_TIME || s->player[i].arrow[j].time_left/FLICKER_RATE%2==0)
	    show_image(arrow_image[i][s->player[i].arrow[j].dir][s->player[i].arrow[j].health], grid2screen(s->player[i].arrow[j].pos));

  /* grid boxes */
  if(s->mode != MODE_INTRO && s->state != POST_GAME)
    for(i=0; i<MAX_PLAYER; i++)
      if(data->player[i].exists)
	if(data->player[i].on_grid)
	  show_image(image[free_for_arrow(s, data->player[i].grid_pos)?GRID_BOX:GRID_BOX_GREY], grid2screen(data->player[i].grid_pos));

  /* vertical walls */
  for(x=0; x<NUM_BLOCKS_X+1; x++)
    for(y=0; y<NUM_BLOCKS_Y; y++)
      if(s->map.vwall[x][y]) {
	p = grid2screen(grid_int_pos(x,y));

	/* this needs to be set every time, see note above */
	rect.w = WALL_THICKNESS;
	rect.h = BLOCK_SIZE;

	rect.x = p.x - rect.w/2;
	rect.y = p.y;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, fence_color.r, fence_color.g, fence_color.b));
	show_image(image[WALLPOST], screen_pos(p.x-image[WALLPOST]->w/2, p.y-image[WALLPOST]->h/2));
	show_image(image[WALLPOST], screen_pos(p.x-image[WALLPOST]->w/2, p.y+BLOCK_SIZE-image[WALLPOST]->h/2));
      }

  /* horiz walls */
  for(x=0; x<NUM_BLOCKS_X; x++)
    for(y=0; y<NUM_BLOCKS_Y+1; y++)
      if(s->map.hwall[x][y]) {
	p = grid2screen(grid_int_pos(x,y));

	/* this needs to be set every time, see note above */
	rect.w = BLOCK_SIZE;
	rect.h = WALL_THICKNESS;

	rect.x = p.x;
	rect.y = p.y - rect.h/2;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, fence_color.r, fence_color.g, fence_color.b));
	show_image(image[WALLPOST], screen_pos(p.x-image[WALLPOST]->w/2, p.y-image[WALLPOST]->h/2));
	show_image(image[WALLPOST], screen_pos(p.x+BLOCK_SIZE-image[WALLPOST]->w/2, p.y-image[WALLPOST]->h/2));
      }

  /* generators */
  for(i=0; i<MAX_GENERATOR; i++)
    if(s->generator[i].exists)
      center_image(image[GENERATOR], grid2screen(s->generator[i].pos));

  /* creatures */
  for(i=0; i<MAX_CREATURE; i++)
    if(s->creature[i].exists) {
      p = gridf2screen(s->creature[i].pos);
      center_image(creature_image(s->creature[i].type), p);
    }

  /* floaters */
  for(i=0; i<MAX_CREATURE; i++)
    if(s->floater[i].exists)
      center_image(creature_image(s->floater[i].type), gridf2screen(s->floater[i].pos));

  /* rockets */
  for(i=0; i<MAX_PLAYER; i++) {
    if(s->player[i].exists) {
      if(s->mode != EVERYBODY_MOVE)
	center_rocket(s, i, grid2screen(s->player[i].rocket));
      else {
	per = 1-((float)s->mode_timer)/EVERYBODY_MOVE_L;

	per = ((pow((per-.5)*2, 3)/2)+.5 + 3*per)/4;

	center_rocket(s, i, gridf2screen(grid_float_pos(s->player[i].old_rocket.x*(1-per) + s->player[i].rocket.x*per, s->player[i].old_rocket.y*(1-per) + s->player[i].rocket.y*per)));
      }
    }
  }

  if(s->mode == CAT_ATTACK)
    for(i=0; i<MAX_PLAYER; i++)
      if(s->player[i].exists && s->player[i].team_leader != s->mode_player)
	center_image(image[CAT], grid2screen(s->player[i].rocket));

  /* score changes */

  for(i=0; i<MAX_PLAYER; i++)
    if(s->player[i].exists && s->player[i].score_change_timer > 0) {
      p = grid2screen(s->player[i].rocket);

      left = p.x + digit_image[0]->w;

      for(j=0,k=1; j<3; j++, k*=10)
	if(abs(s->player[i].score_change)/k > 0 || j==0) {
	  show_image(digit_image[abs(s->player[i].score_change)/k%10], screen_pos(left + digit_image[0]->w - digit_image[abs(s->player[i].score_change)/k%10]->w, p.y));
	  left -= digit_image[0]->w;
	}

      if(s->player[i].score_change > 0)
	show_image(digit_image[PLUS], screen_pos(left, p.y));
      if(s->player[i].score_change < 0)
	show_image(digit_image[MINUS], screen_pos(left, p.y));
    }

  /* interface stuff */


  for(i=0; i<MAX_PLAYER; i++)
    if(s->player[i].exists && s->player[i].team_leader == i)
      player_rank[i] = -1;
  
  for(j=1; j<=MAX_PLAYER; j++) {
    /* find max player */
    max_score=-1;
    max_player=-1;
    for(i=0; i<MAX_PLAYER; i++)
      if(s->player[i].exists && s->player[i].team_leader == i)
	if(player_rank[i] == -1 && s->player[i].score > max_score) {
	  max_score = s->player[i].score;
	  max_player = i;
	}

    if(max_player != -1)
      player_rank[max_player] = j;

    for(i=0; i<MAX_PLAYER; i++)
      if(s->player[i].exists && s->player[i].team_leader == i)
	if(player_rank[i] == -1 && s->player[i].score == max_score)
	  player_rank[i] = j;
  }

#define SCORE_BOX_WIDTH 5

  /* score */
  for(i=0; i<MAX_PLAYER; i++)
    if(s->player[i].exists) {
      rect.x = 10+(BLOCK_SIZE*NUM_BLOCKS_X)/MAX_PLAYER*i;
      rect.y = screen->h-digit_image[0]->h-5;
      rect.w = digit_image[0]->w*3;
      rect.h = digit_image[0]->h;
      rect.x-=SCORE_BOX_WIDTH; rect.y-=SCORE_BOX_WIDTH;
      rect.w += SCORE_BOX_WIDTH*2; rect.h += SCORE_BOX_WIDTH*2;
      SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, player_color[i].r, player_color[i].g, player_color[i].b));
      rect.x+=SCORE_BOX_WIDTH; rect.y+=SCORE_BOX_WIDTH;
      rect.w -= SCORE_BOX_WIDTH*2; rect.h -= SCORE_BOX_WIDTH*2;
      SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, black.r, black.g, black.b));

      rect.x -= SCORE_BOX_WIDTH;
      rect.y -= SCORE_BOX_WIDTH;
      rect.w += SCORE_BOX_WIDTH*2;
      rect.h += SCORE_BOX_WIDTH*2;

      for(j=0,k=1; j<3; j++,k*=10) {
	show_image(digit_image[s->player[s->player[i].team_leader].score/k%10], screen_pos(10+(BLOCK_SIZE*NUM_BLOCKS_X)/MAX_PLAYER*i+digit_image[0]->w*(3-j)-digit_image[s->player[s->player[i].team_leader].score/k%10]->w, screen->h-digit_image[0]->h-5));
      }

      /* places */

      temp_image = NULL;

      switch(player_rank[s->player[i].team_leader]) {
      case 1: temp_image = image[PLACE_1]; break;
      case 2: temp_image = image[PLACE_2]; break;
      case 3: temp_image = image[PLACE_3]; break;
      case 4: temp_image = image[PLACE_4]; break;
      }

      if(s->state != NOT_STARTED)
	if(temp_image != NULL)
	  show_image(temp_image, screen_pos(rect.x+rect.w-image[PLACE_1]->w/2, rect.y-image[PLACE_1]->h/2));

    }

  /* clock */
  show_image(digit_image[s->clock/1000%60%10], screen_pos(screen->w-digit_image[0]->w-5, screen->h-digit_image[0]->h-5));
  show_image(digit_image[s->clock/1000%60/10%10], screen_pos(screen->w-digit_image[0]->w*2-5, screen->h-digit_image[0]->h-5));
  show_image(digit_image[s->clock/1000/60%10], screen_pos(screen->w-digit_image[0]->w*3-5, screen->h-digit_image[0]->h-5));

  if(s->mode == MODE_INTRO || s->state == POST_GAME) {
    /* mode intro box */
    p = gridf2screen(grid_float_pos(0.5, 3.5));
    
    rect.x = p.x;
    rect.w = BLOCK_SIZE*11;

    rect.y = p.y;
    rect.h = BLOCK_SIZE*2;
    
    SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, box_border_color.r, box_border_color.g, box_border_color.b));

    rect.x += 2;
    rect.w -= 2*2;
    rect.y += 2;
    rect.h -= 2*2;

    i = get_winner(s);

    if(s->state == POST_GAME && i>=0)
      SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, player_color[i].r/2, player_color[i].g/2, player_color[i].b/2));
    else
      SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, box_color.r, box_color.g, box_color.b));

    if(s->state == POST_GAME) {
      if(i>=0)
	center_text(rect.x+rect.w/2, rect.y+rect.h/2, "Winner!");
      else
	center_text(rect.x+rect.w/2, rect.y+rect.h/2, "Tie!");
    }
    else
      center_text(rect.x+rect.w/2, rect.y+rect.h/2, mode_string(s->next_mode));


  }
  else {
    /* pointers */
    for(i=0; i<MAX_PLAYER; i++) {
      if(data->player[i].exists)
	ShowImage(player_pointer_image(i), data->player[i].pointer.x-2, data->player[i].pointer.y-2);
    }
  }


  SDL_Flip(screen);
}



void make_grid_box(SDL_Surface *s, SDL_Color c)
{
  SDL_Rect rect;

  rect.x = 0;
  rect.y = 0;
  rect.w = BLOCK_SIZE;
  rect.h = BLOCK_SIZE;
  
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, 0, 0, 0));

  rect.x = 0;
  rect.y = 0;
  rect.w = BLOCK_SIZE*2/5;
  rect.h = GRID_BOX_THICKNESS;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
  rect.x = BLOCK_SIZE*3/5;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
  rect.x = 0;
  
  rect.y = BLOCK_SIZE-GRID_BOX_THICKNESS;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
  rect.x = BLOCK_SIZE*3/5;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));

  rect.x = 0;
  rect.y = 0;
  rect.w = GRID_BOX_THICKNESS;
  rect.h = BLOCK_SIZE*2/5;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
  rect.y = BLOCK_SIZE*3/5;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
  rect.y = 0;

  rect.x = BLOCK_SIZE-GRID_BOX_THICKNESS;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
  rect.y = BLOCK_SIZE*3/5;
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));

  /*  SDL_SetAlpha(s, SDL_SRCALPHA, 200); */
}

void arrow_rect(SDL_Surface *s, direction dir, SDL_Rect rect, SDL_Color c)
{
  int temp;
  if(dir==down) {
    rect.y = BLOCK_SIZE-rect.y-rect.h;
  }
  if(dir==right || dir==left) {
    temp = rect.x;
    rect.x = rect.y;
    rect.y = temp;
    
    temp = rect.w;
    rect.w = rect.h;
    rect.h = temp;
  }
  if(dir==right) {
    rect.x = BLOCK_SIZE-rect.x-rect.w;
  }
  SDL_FillRect(s, &rect, SDL_MapRGB(s->format, c.r, c.g, c.b));
}

void make_arrow(SDL_Surface *s, direction dir, SDL_Color c, int health)
{
  int i;
  int w;
  int thisw;
  int h,hbox,btop;

#define SHRINK_FACTOR 5
#define BOX_SHRINK_FACTOR 1/10

  SDL_Rect rect = {0, 0, 0, 0};
  rect.w = rect.h = BLOCK_SIZE;

  arrow_rect(s, dir, rect, black);

  rect.x = rect.y = BLOCK_SIZE*BOX_SHRINK_FACTOR*(ARROW_TOUGHNESS-1-health);
  rect.w = rect.h = BLOCK_SIZE-BLOCK_SIZE*2*BOX_SHRINK_FACTOR*(ARROW_TOUGHNESS-1-health);

  arrow_rect(s, dir, rect, c);

  w = 3;
  h = BLOCK_SIZE*8/14*(health+SHRINK_FACTOR)/(SHRINK_FACTOR+1);
  hbox = BLOCK_SIZE/4*(health+SHRINK_FACTOR)/(SHRINK_FACTOR+1);
  btop = (BLOCK_SIZE-h-hbox)/2;

  for(i=0; i<h; i++) {
    thisw = w/3 /2*2;
    /*    if(i==h-1)
	  thisw-=2;*/
    rect.x = BLOCK_SIZE/2-thisw/2;
    rect.y = i+btop;
    rect.w = thisw;
    rect.h = 1;
    
    arrow_rect(s, dir, rect, arrow_color);
    w += 3;
  }
  
  rect.y = rect.h + rect.y;
  rect.h = hbox;
  rect.x = BLOCK_SIZE*3/8;
  rect.w = BLOCK_SIZE/4;
  
  arrow_rect(s, dir, rect, arrow_color);
}

int load_images()
{
  int i,j,k;
  SDL_Rect rect;
  char buff[100];

  printf("Loading images...\n");
  
  for(i=0; i<MAX_IMAGE; i++) {
    for(j=0; j<MAX_IMAGE; j++)
      if(image_info[j].id == i)
	break;
    if(j!=MAX_IMAGE) { /* if there is something to load for it */

      sprintf(&buff[0], "images/%s", image_info[j].file);
    
      image[i] = LoadImage(&buff[0], image_info[j].trans);
      if(image[i]==NULL) {
	printf("Couldn't load %s\n", image_info[j].file);
	return 0;
      }
      
      if(image_info[j].alpha)
	SDL_SetAlpha(image[i], SDL_SRCALPHA, image_info[j].alpha);
    }
  }

  image[WALLPOST] = SDL_CreateRGBSurface(SDL_HWSURFACE|SDL_SRCCOLORKEY, WALLPOST_SIZE, WALLPOST_SIZE, 16, 0, 0, 0, 0);
  SDL_SetColorKey(image[WALLPOST], (SDL_SRCCOLORKEY|SDL_RLEACCEL), SDL_MapRGB(image[WALLPOST]->format, 0, 0, 1));
  
  for(i=0; i<WALLPOST_SIZE/2; i++) {
    rect.w = i*2+2;
    rect.x = WALLPOST_SIZE/2-i-1;
    rect.h = 1;
    rect.y = i;
    SDL_FillRect(image[WALLPOST], &rect, SDL_MapRGB(image[WALLPOST]->format, fence_post_color.r, fence_post_color.g, fence_post_color.b));
  }

  for(i=0; i<=WALLPOST_SIZE/2; i++) {
    rect.w = (WALLPOST_SIZE/2-i)*2;
    rect.x = i;
    rect.h = 1;
    rect.y = WALLPOST_SIZE/2+i;
    SDL_FillRect(image[WALLPOST], &rect, SDL_MapRGB(image[WALLPOST]->format, fence_post_color.r, fence_post_color.g, fence_post_color.b));
  }

  image[GRID_BOX] = SDL_CreateRGBSurface(SDL_HWSURFACE|SDL_SRCCOLORKEY, BLOCK_SIZE, BLOCK_SIZE, 16, 0, 0, 0, 0);
  SDL_SetColorKey(image[GRID_BOX], (SDL_SRCCOLORKEY|SDL_RLEACCEL), SDL_MapRGB(image[WALLPOST]->format, 0, 0, 0));

  /* create grid box */
  make_grid_box(image[GRID_BOX], grid_box_color);

  image[GRID_BOX_GREY] = SDL_CreateRGBSurface(SDL_HWSURFACE|SDL_SRCCOLORKEY, BLOCK_SIZE, BLOCK_SIZE, 16, 0, 0, 0, 0);
  SDL_SetColorKey(image[GRID_BOX_GREY], (SDL_SRCCOLORKEY|SDL_RLEACCEL), SDL_MapRGB(image[WALLPOST]->format, 0, 0, 0));
  make_grid_box(image[GRID_BOX_GREY], grid_box_grey_color);

  /* create arrows */
  for(i=0; i<MAX_PLAYER; i++)
    for(j=0; j<NUM_DIR; j++)
      for(k=0; k<ARROW_TOUGHNESS; k++) {
	arrow_image[i][j][k] = SDL_CreateRGBSurface(SDL_HWSURFACE, BLOCK_SIZE, BLOCK_SIZE, 16, 0, 0, 0, 0);
	SDL_SetColorKey(arrow_image[i][j][k], (SDL_SRCCOLORKEY|SDL_RLEACCEL), SDL_MapRGB(arrow_image[i][j][k]->format, 0, 0, 0));
	make_arrow(arrow_image[i][j][k], j, player_color[i], k);
    }

  for(i=0; i<12; i++) {
    sprintf(&buff[0], "./images/%d.gif", i);
    digit_image[i] = LoadImage(&buff[0], 0);
    
    if(digit_image[i] == NULL) {
      fprintf(stderr, "Couldn't load digit %d.\n", i);
      return 0;
    }

    SDL_SetColorKey(digit_image[i], (SDL_SRCCOLORKEY|SDL_RLEACCEL), SDL_MapRGB(digit_image[i]->format, 0, 0, 0));
  }

  font = LoadImage("./images/font2.bmp", 0);
  InitFont(font);

  return 1;
}

void unload_images()
{
  int i,j,k;

  for(i=0; i<MAX_IMAGE; i++)
    SDL_FreeSurface(image[i]);

  for(i=0; i<MAX_PLAYER; i++)
    for(j=0; j<NUM_DIR; j++)
      for(k=0; k<ARROW_TOUGHNESS; k++)
	SDL_FreeSurface(arrow_image[i][j][k]);
  
  for(i=0; i<12; i++)
    SDL_FreeSurface(digit_image[i]);
}

void output_screenshot()
{
  char *l = "shot.bmp";
  SDL_SaveBMP(screen, l);
  printf("Screenshot saved to %s.\n", l);
}

int output_init(int fullscreen)
{

  set_mode(SCREEN_WIDTH, SCREEN_HEIGHT, fullscreen);

  SDL_ShowCursor(0);

  /* images are loaded by set_mode */

  if(!audio_init()) {
    printf("Audio init failed.\n");
    return 0;
  }

  return 1;
}


int output_exit()
{
  unload_images();
  audio_exit();
  return 1;
}
  
